// include
//--------------------------
#include <windows.h>
#include <assert.h>
#include "x86.h"


// defines
//-----------------------------
#define BX_CPU_LEVEL             5
#define BX_SEGMENT_REG           10
#define BX_GENERAL_8BIT_REG      11
#define BX_GENERAL_16BIT_REG     12
#define BX_GENERAL_32BIT_REG     13
#define BX_NO_REG_TYPE           14
#define BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm) \
   { \
      mod      = (mod_rm_byte >> 6) & 0x03; \
      opcode   = (mod_rm_byte >> 3) & 0x07; \
      rm       =  mod_rm_byte & 0x07; \
   }

// globals
//-----------------------------
bool                 valid = true;
bool                 db_32bit_opsize;
bool                 db_32bit_addrsize;
bool                 db_rep_prefix;
bool                 db_repne_prefix;
const unsigned char* instruction_begin;   // keep track of where instruction starts
const unsigned char* instruction;         // for fetching of next byte of instruction


// local prototypes
//-----------------------------
unsigned char  fetch_byte(void);
unsigned char  peek_byte(void);
unsigned short fetch_word(void);
unsigned int   fetch_dword(void);

void decode_exgx(unsigned modrm_reg_type, unsigned reg_type);
void decode_gxex(unsigned reg_type, unsigned modrm_reg_type);
void decode_ex(unsigned char modrm, unsigned modrm_reg_type);

void EbGb(void);
void EvGv(void);
void GbEb(void);
void GvEv(void);
void Ew(void);
void Ms(void);
void GvEw(void);
void RdCd(void);
void RdDd(void);
void CdRd(void);
void DdRd(void);
void RdTd(void);
void TdRd(void);
void Jv(void);
void Eb(void);
void XBTS(void);
void IBTS(void);
void Mp(void);
void EvIb(void);
void GvEb(void);
void GvMa(void);
void EwRw(void);
void Iv(void);
void Ib(void);
void YbDX(void);
void YvDX(void);
void DXXb(void);
void DXXv(void);
void Jb(void);
void EbIb(void);
void EvIv(void);
void EwSw(void);
void GvM(void);
void SwEw(void);
void Ev(void);

void eAX(void);
void eCX(void);
void eDX(void);
void eBX(void);
void eSP(void);
void eBP(void);
void eSI(void);
void eDI(void);

void Ap(void);
void ALOb(void);
void eAXOv(void);
void ObAL(void);
void OveAX(void);
void XbYb(void);
void XvYv(void);
void YbAL(void);
void YveAX(void);
void ALXb(void);
void eAXXv(void);

void eSPIv(void);
void eBPIv(void);
void Iw(void);
void GvMp(void);
void Eb1(void);
void Ev1(void);
void Es(void);
void ST_STi(void);
void Ea(void);
void STi(void);
void Et(void);
void Ed(void);
void El(void);
void STi_ST(void);
void Eq(void);
void Av(void);
void eAXEv(void);
void Ep(void);


// x86GetInstrBufferLength
//-----------------------------
bool x86GetInstrBufferLength( const unsigned char* instr, unsigned int min_length, unsigned int& length )
{
   for(length=0; length<min_length; )
   {
      unsigned int   temp_length;
      if(!x86GetInstrLength(instr+length, temp_length))
         return false;
      length += temp_length;
   }
   return true;
}


// x86GetInstrLength
//-----------------------------
bool x86GetInstrLength( const unsigned char* instr, unsigned int& length )
{
   int               byte_count;
   unsigned char     next_byte;
   unsigned char     mod_rm_byte, mod, opcode, rm;

   valid = true;
   db_32bit_opsize = db_32bit_addrsize = true;
   instruction_begin = instruction = instr;

   db_rep_prefix = 0;
   db_repne_prefix = 0;

   // On the 386 and above, instructions must be a maximum of 15 bytes long.
   // this means redundant prefix codes can put the byte count over 15 and
   // cause an illegal instruction.
   for (byte_count=0; byte_count<15; byte_count++)
   {
      next_byte = fetch_byte();
      switch (next_byte)
      {
      case 0x00: EbGb(); goto done;
      case 0x01: EvGv(); goto done;
      case 0x02: GbEb(); goto done;
      case 0x03: GvEv(); goto done;
      case 0x04: Ib(); goto done;
      case 0x05: eAX(); Iv(); goto done;
      case 0x06: goto done;
      case 0x07: goto done;
      case 0x08: EbGb(); goto done;
      case 0x09: EvGv(); goto done;
      case 0x0A: GbEb(); goto done;
      case 0x0B: GvEv(); goto done;
      case 0x0C: Ib(); goto done;
      case 0x0D: eAX(); Iv(); goto done;
      case 0x0E: goto done;
      case 0x0F: // 2-byte escape
         next_byte = fetch_byte();
         if ( ++byte_count >= 15 )
         {
            valid = false;
            goto done;
         }
         switch (next_byte)
         {
         case 0x00: // Group 6
            mod_rm_byte = peek_byte();
            BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
            switch (opcode)
            {
            case 0x00: Ew(); goto done;
            case 0x01: Ew(); goto done;
            case 0x02: Ew(); goto done;
            case 0x03: Ew(); goto done;
            case 0x04: Ew(); goto done;
            case 0x05: Ew(); goto done;
            case 0x06: valid = false; goto done;
            case 0x07: valid = false; goto done;
            default: valid = false; goto done;
            }
         case 0x01: // Group 7
            mod_rm_byte = peek_byte();
            BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
            switch (opcode)
            {
            case 0x00: Ms(); goto done;
            case 0x01: Ms(); goto done;
            case 0x02: Ms(); goto done;
            case 0x03: Ms(); goto done;
            case 0x04: Ew(); goto done;
            case 0x05: valid = false; goto done;
            case 0x06: Ew(); goto done;
            case 0x07: valid = false; goto done;
            default: valid = false; goto done;
            }
         case 0x02: GvEw(); goto done;
         case 0x03: GvEw(); goto done;
         case 0x04: valid = false; goto done;
         case 0x05: valid = false; goto done;
         case 0x06: goto done;
         case 0x07: valid = false; goto done;
         case 0x08: goto done;
         case 0x09: goto done;
         case 0x0A:
         case 0x0B:
         case 0x0C:
         case 0x0D:
         case 0x0E:
         case 0x0F:

         case 0x10:
         case 0x11:
         case 0x12:
         case 0x13:
         case 0x14:
         case 0x15:
         case 0x16:
         case 0x17:
         case 0x18:
         case 0x19:
         case 0x1A:
         case 0x1B:
         case 0x1C:
         case 0x1D:
         case 0x1E:
         case 0x1F: valid = false; goto done;

         case 0x20: RdCd(); goto done;
         case 0x21: RdDd(); goto done;
         case 0x22: CdRd(); goto done;
         case 0x23: DdRd(); goto done;
         case 0x24: RdTd(); goto done;
         case 0x25: valid = false; goto done;
         case 0x26: TdRd(); goto done;
         case 0x27:
         case 0x28:
         case 0x29:
         case 0x2A:
         case 0x2B:
         case 0x2C:
         case 0x2D:
         case 0x2E:
         case 0x2F:

         case 0x30:
         case 0x31:
         case 0x32:
         case 0x33:
         case 0x34:
         case 0x35:
         case 0x36:
         case 0x37:
         case 0x38:
         case 0x39:
         case 0x3A:
         case 0x3B:
         case 0x3C:
         case 0x3D:
         case 0x3E:
         case 0x3F:

         case 0x40:
         case 0x41:
         case 0x42:
         case 0x43:
         case 0x44:
         case 0x45:
         case 0x46:
         case 0x47:
         case 0x48:
         case 0x49:
         case 0x4A:
         case 0x4B:
         case 0x4C:
         case 0x4D:
         case 0x4E:
         case 0x4F:

         case 0x50:
         case 0x51:
         case 0x52:
         case 0x53:
         case 0x54:
         case 0x55:
         case 0x56:
         case 0x57:
         case 0x58:
         case 0x59:
         case 0x5A:
         case 0x5B:
         case 0x5C:
         case 0x5D:
         case 0x5E:
         case 0x5F:

         case 0x60:
         case 0x61:
         case 0x62:
         case 0x63:
         case 0x64:
         case 0x65:
         case 0x66:
         case 0x67:
         case 0x68:
         case 0x69:
         case 0x6A:
         case 0x6B:
         case 0x6C:
         case 0x6D:
         case 0x6E:
         case 0x6F:

         case 0x70:
         case 0x71:
         case 0x72:
         case 0x73:
         case 0x74:
         case 0x75:
         case 0x76:
         case 0x77:
         case 0x78:
         case 0x79:
         case 0x7A:
         case 0x7B:
         case 0x7C:
         case 0x7D:
         case 0x7E:
         case 0x7F: valid = false; goto done;

         case 0x80:
         case 0x81:
         case 0x82:
         case 0x83:
         case 0x84:
         case 0x85:
         case 0x86:
         case 0x87:
         case 0x88:
         case 0x89:
         case 0x8A:
         case 0x8B:
         case 0x8C:
         case 0x8D:
         case 0x8E:
         case 0x8F: Jv(); goto done;

         case 0x90:
         case 0x91:
         case 0x92:
         case 0x93:
         case 0x94:
         case 0x95:
         case 0x96:
         case 0x97:
         case 0x98:
         case 0x99:
         case 0x9A:
         case 0x9B:
         case 0x9C:
         case 0x9D:
         case 0x9E:
         case 0x9F: Eb(); goto done;

         case 0xA0: goto done;
         case 0xA1: goto done;
         case 0xA2: valid = false; goto done;
         case 0xA3: EvGv(); goto done;
         case 0xA4: EvGv(); Ib(); goto done;
         case 0xA5: EvGv(); goto done;
         case 0xA6: XBTS(); goto done;
         case 0xA7: IBTS(); goto done;
         case 0xA8: goto done;
         case 0xA9: goto done;
         case 0xAA: valid = false; goto done;
         case 0xAB: EvGv(); goto done;
         case 0xAC: EvGv(); Ib(); goto done;
         case 0xAD: EvGv(); goto done;
         case 0xAE: valid = false; goto done;
         case 0xAF: GvEv(); goto done;

         case 0xB0: EbGb(); goto done;
         case 0xB1: EvGv(); goto done;
         case 0xB2: Mp(); goto done;
         case 0xB3: EvGv(); goto done;
         case 0xB4: Mp(); goto done;
         case 0xB5: Mp(); goto done;
         case 0xB6: GvEb(); goto done;
         case 0xB7: GvEw(); goto done;
         case 0xB8: valid = false; goto done;
         case 0xB9: valid = false; goto done;
         case 0xBA: // Group 8 Ev,Ib
            mod_rm_byte = peek_byte();
            BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
            switch (opcode)
            {
               case 0x00:
               case 0x01:
               case 0x02:
               case 0x03: valid = false; goto done;
               case 0x04: EvIb(); goto done;
               case 0x05: EvIb(); goto done;
               case 0x06: EvIb(); goto done;
               case 0x07: EvIb(); goto done;
               default: valid = false; goto done;
            }

         case 0xBB: EvGv(); goto done;
         case 0xBC: GvEv(); goto done;
         case 0xBD: GvEv(); goto done;
         case 0xBE: GvEb(); goto done;
         case 0xBF: GvEw(); goto done;

         case 0xC0: EbGb(); goto done;
         case 0xC1: EvGv(); goto done;
         case 0xC2:
         case 0xC3:
         case 0xC4:
         case 0xC5:
         case 0xC6:
         case 0xC7: valid = false; goto done;
         case 0xC8: eAX(); goto done;
         case 0xC9: eCX(); goto done;
         case 0xCA: eDX(); goto done;
         case 0xCB: eBX(); goto done;
         case 0xCC: eSP(); goto done;
         case 0xCD: eBP(); goto done;
         case 0xCE: eSI(); goto done;
         case 0xCF: eDI(); goto done;

         case 0xD0:
         case 0xD1:
         case 0xD2:
         case 0xD3:
         case 0xD4:
         case 0xD5:
         case 0xD6:
         case 0xD7:
         case 0xD8:
         case 0xD9:
         case 0xDA:
         case 0xDB:
         case 0xDC:
         case 0xDD:
         case 0xDE:
         case 0xDF:

         case 0xE0:
         case 0xE1:
         case 0xE2:
         case 0xE3:
         case 0xE4:
         case 0xE5:
         case 0xE6:
         case 0xE7:
         case 0xE8:
         case 0xE9:
         case 0xEA:
         case 0xEB:
         case 0xEC:
         case 0xED:
         case 0xEE:
         case 0xEF:

         case 0xF0:
         case 0xF1:
         case 0xF2:
         case 0xF3:
         case 0xF4:
         case 0xF5:
         case 0xF6:
         case 0xF7:
         case 0xF8:
         case 0xF9:
         case 0xFA:
         case 0xFB:
         case 0xFC:
         case 0xFD:
         case 0xFE:
         case 0xFF:

         default: valid = false; goto done;
         }

      case 0x10: EbGb(); goto done;
      case 0x11: EvGv(); goto done;
      case 0x12: GbEb(); goto done;
      case 0x13: GvEv(); goto done;
      case 0x14: Ib(); goto done;
      case 0x15: eAX(); Iv(); goto done;
      case 0x16: goto done;
      case 0x17: goto done;
      case 0x18: EbGb(); goto done;
      case 0x19: EvGv(); goto done;
      case 0x1A: GbEb(); goto done;
      case 0x1B: GvEv(); goto done;
      case 0x1C: Ib(); goto done;
      case 0x1D: eAX(); Iv(); goto done;
      case 0x1E: goto done;
      case 0x1F: goto done;

      case 0x20: EbGb(); goto done;
      case 0x21: EvGv(); goto done;
      case 0x22: GbEb(); goto done;
      case 0x23: GvEv(); goto done;
      case 0x24: Ib(); goto done;
      case 0x25: eAX(); Iv(); goto done;
      case 0x26:
         break;
      case 0x27: goto done;
      case 0x28: EbGb(); goto done;
      case 0x29: EvGv(); goto done;
      case 0x2A: GbEb(); goto done;
      case 0x2B: GvEv(); goto done;
      case 0x2C: Ib(); goto done;
      case 0x2D: eAX(); Iv(); goto done;
      case 0x2E:
         break;
      case 0x2F: goto done;

      case 0x30: EbGb(); goto done;
      case 0x31: EvGv(); goto done;
      case 0x32: GbEb(); goto done;
      case 0x33: GvEv(); goto done;
      case 0x34: Ib(); goto done;
      case 0x35: eAX(); Iv(); goto done;
      case 0x36:
         break;
      case 0x37: goto done;
      case 0x38: EbGb(); goto done;
      case 0x39: EvGv(); goto done;
      case 0x3A: GbEb(); goto done;
      case 0x3B: GvEv(); goto done;
      case 0x3C: Ib(); goto done;
      case 0x3D: eAX(); Iv(); goto done;
      case 0x3E:
         break;
      case 0x3F: goto done;

      case 0x40: eAX(); goto done;
      case 0x41: eCX(); goto done;
      case 0x42: eDX(); goto done;
      case 0x43: eBX(); goto done;
      case 0x44: eSP(); goto done;
      case 0x45: eBP(); goto done;
      case 0x46: eSI(); goto done;
      case 0x47: eDI(); goto done;
      case 0x48: eAX(); goto done;
      case 0x49: eCX(); goto done;
      case 0x4A: eDX(); goto done;
      case 0x4B: eBX(); goto done;
      case 0x4C: eSP(); goto done;
      case 0x4D: eBP(); goto done;
      case 0x4E: eSI(); goto done;
      case 0x4F: eDI(); goto done;

      case 0x50: eAX(); goto done;
      case 0x51: eCX(); goto done;
      case 0x52: eDX(); goto done;
      case 0x53: eBX(); goto done;
      case 0x54: eSP(); goto done;
      case 0x55: eBP(); goto done;
      case 0x56: eSI(); goto done;
      case 0x57: eDI(); goto done;
      case 0x58: eAX(); goto done;
      case 0x59: eCX(); goto done;
      case 0x5A: eDX(); goto done;
      case 0x5B: eBX(); goto done;
      case 0x5C: eSP(); goto done;
      case 0x5D: eBP(); goto done;
      case 0x5E: eSI(); goto done;
      case 0x5F: eDI(); goto done;

      case 0x60: goto done;
      case 0x61: goto done;
      case 0x62: GvMa(); goto done;
      case 0x63: EwRw(); goto done;
      case 0x64:
         break;
      case 0x65:
         break;
      case 0x66:
         db_32bit_opsize = !db_32bit_opsize;
         break;
      case 0x67:
         db_32bit_addrsize = !db_32bit_addrsize;
         break;
      case 0x68: Iv(); goto done;
      case 0x69: GvEv(); Iv(); goto done;
      case 0x6A: Ib(); goto done;
      case 0x6B: GvEv(); Ib(); goto done;
      case 0x6C: YbDX(); goto done;
      case 0x6D: YvDX(); goto done;
      case 0x6E: DXXb(); goto done;
      case 0x6F: DXXv(); goto done;

      case 0x70:
      case 0x71:
      case 0x72:
      case 0x73:
      case 0x74:
      case 0x75:
      case 0x76:
      case 0x77:
      case 0x78:
      case 0x79:
      case 0x7A:
      case 0x7B:
      case 0x7C:
      case 0x7D:
      case 0x7E:
      case 0x7F: Jb(); goto done;

      case 0x80: // Immdediate Grp 1 EbIb
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         switch (opcode)
         {
         case 0x00: EbIb(); goto done;
         case 0x01: EbIb();  goto done;
         case 0x02: EbIb(); goto done;
         case 0x03: EbIb(); goto done;
         case 0x04: EbIb(); goto done;
         case 0x05: EbIb(); goto done;
         case 0x06: EbIb(); goto done;
         case 0x07: EbIb(); goto done;
         default: valid = false; goto done;
         }
         break;

      case 0x81: // Immdediate Grp 1 EvIv
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         switch (opcode)
         {
         case 0x00: EvIv(); goto done;
         case 0x01: EvIv(); goto done;
         case 0x02: EvIv(); goto done;
         case 0x03: EvIv(); goto done;
         case 0x04: EvIv(); goto done;
         case 0x05: EvIv(); goto done;
         case 0x06: EvIv(); goto done;
         case 0x07: EvIv(); goto done;
         default: valid = false; goto done;
         }

      case 0x82: valid = false; goto done;
      case 0x83: // Immdediate Grp 1 EvIb
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         switch (opcode)
         {
         case 0x00: EvIb(); goto done;
         case 0x01: EvIb(); goto done;
         case 0x02: EvIb(); goto done;
         case 0x03: EvIb(); goto done;
         case 0x04: EvIb(); goto done;
         case 0x05: EvIb(); goto done;
         case 0x06: EvIb(); goto done;
         case 0x07: EvIb(); goto done;
         default: valid = false; goto done;
         }

      case 0x84: EbGb(); goto done;
      case 0x85: EvGv(); goto done;
      case 0x86: EbGb(); goto done;
      case 0x87: EvGv(); goto done;
      case 0x88: EbGb(); goto done;
      case 0x89: EvGv(); goto done;
      case 0x8A: GbEb(); goto done;
      case 0x8B: GvEv(); goto done;
      case 0x8C: EwSw(); goto done;
      case 0x8D: GvM(); goto done;
      case 0x8E: SwEw(); goto done;
      case 0x8F: Ev(); goto done;

      case 0x90: goto done;
      case 0x91: eCX(); eAX(); goto done;
      case 0x92: eDX(); eAX(); goto done;
      case 0x93: eBX(); eAX(); goto done;
      case 0x94: eSP(); eAX(); goto done;
      case 0x95: eBP(); eAX(); goto done;
      case 0x96: eSI(); eAX(); goto done;
      case 0x97: eDI(); eAX(); goto done;
      case 0x98: goto done;
      case 0x99: goto done;
      case 0x9A: Ap(); goto done;
      case 0x9B: goto done;
      case 0x9C: goto done;
      case 0x9D: goto done;
      case 0x9E: goto done;
      case 0x9F: goto done;

      case 0xA0: ALOb(); goto done;
      case 0xA1: eAXOv(); goto done;
      case 0xA2: ObAL(); goto done;
      case 0xA3: OveAX(); goto done;
      case 0xA4: XbYb(); goto done;
      case 0xA5: 
         if (db_32bit_opsize)
            XvYv();
         else
            XvYv();
         goto done;
      case 0xA6: XbYb(); goto done;
      case 0xA7: XvYv(); goto done;
      case 0xA8: Ib(); goto done;
      case 0xA9: eAX(); Iv(); goto done;
      case 0xAA: YbAL(); goto done;
      case 0xAB: YveAX(); goto done;
      case 0xAC: ALXb(); goto done;
      case 0xAD: eAXXv(); goto done;
      case 0xAE: ALXb(); goto done;
      case 0xAF: eAXXv(); goto done;

      case 0xB0: Ib(); goto done;
      case 0xB1: Ib(); goto done;
      case 0xB2: Ib(); goto done;
      case 0xB3: Ib(); goto done;
      case 0xB4: Ib(); goto done;
      case 0xB5: Ib(); goto done;
      case 0xB6: Ib(); goto done;
      case 0xB7: Ib(); goto done;
      case 0xB8: eAX(); Iv(); goto done;
      case 0xB9: eCX(); Iv(); goto done;
      case 0xBA: eDX(); Iv(); goto done;
      case 0xBB: eBX(); Iv(); goto done;
      case 0xBC: eSP(); Iv(); goto done;
      case 0xBD: eBP(); Iv(); goto done;
      case 0xBE: eSI(); Iv(); goto done;
      case 0xBF: eDI(); Iv(); goto done;

      case 0xC0: // Group 2 Eb,Ib
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         switch (opcode)
         {
         case 0x00: EbIb(); goto done;
         case 0x01: EbIb(); goto done;
         case 0x02: EbIb(); goto done;
         case 0x03: EbIb(); goto done;
         case 0x04: EbIb(); goto done;
         case 0x05: EbIb(); goto done;
         case 0x06: valid = false; goto done;
         case 0x07: EbIb(); goto done;
         default: valid = false; goto done;
         }

      case 0xC1: // Group 2 Ev,Ib
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         switch (opcode)
         {
         case 0x00: EvIb(); goto done;
         case 0x01: EvIb(); goto done;
         case 0x02: EvIb(); goto done;
         case 0x03: EvIb(); goto done;
         case 0x04: EvIb(); goto done;
         case 0x05: EvIb(); goto done;
         case 0x06: valid = false; goto done;
         case 0x07: EvIb(); goto done;
         default: valid = false; goto done;
         }

      case 0xC2: Iw(); goto done;
      case 0xC3: goto done;
      case 0xC4: GvMp(); goto done;
      case 0xC5: GvMp(); goto done;
      case 0xC6: EbIb(); goto done;
      case 0xC7: EvIv(); goto done;
      case 0xC8: Iw(); Ib(); goto done;
      case 0xC9: goto done;
      case 0xCA: Iw(); goto done;
      case 0xCB: goto done;
      case 0xCC: goto done;
      case 0xCD: Ib(); goto done;
      case 0xCE: goto done;
      case 0xCF: goto done;

      case 0xD0: // Group 2 Eb,1
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         switch (opcode)
         {
         case 0x00: Eb1(); goto done;
         case 0x01: Eb1(); goto done;
         case 0x02: Eb1(); goto done;
         case 0x03: Eb1(); goto done;
         case 0x04: Eb1(); goto done;
         case 0x05: Eb1(); goto done;
         case 0x06: valid = false; goto done;
         case 0x07: Eb1(); goto done;
         default: valid = false; goto done;
         }

      case 0xD1: // group2 Ev,1
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         switch (opcode)
         {
         case 0x00: Ev1(); goto done;
         case 0x01: Ev1(); goto done;
         case 0x02: Ev1(); goto done;
         case 0x03: Ev1(); goto done;
         case 0x04: Ev1(); goto done;
         case 0x05: Ev1(); goto done;
         case 0x06: valid = false; goto done;
         case 0x07: Ev1(); goto done;
         default: valid = false; goto done;
         }

      case 0xD2: // group2 Eb,CL
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         switch (opcode)
         {
         case 0x00: Eb(); goto done;
         case 0x01: Eb(); goto done;
         case 0x02: Eb(); goto done;
         case 0x03: Eb(); goto done;
         case 0x04: Eb(); goto done;
         case 0x05: Eb(); goto done;
         case 0x06: valid = false; goto done;
         case 0x07: Eb(); goto done;
         default: valid = false; goto done;
         }

      case 0xD3: // group2 Ev,CL
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         switch (opcode)
         {
         case 0x00: Ev(); goto done;
         case 0x01: Ev(); goto done;
         case 0x02: Ev(); goto done;
         case 0x03: Ev(); goto done;
         case 0x04: Ev(); goto done;
         case 0x05: Ev(); goto done;
         case 0x06: valid = false; goto done;
         case 0x07: Ev(); goto done;
         default: valid = false; goto done;
         }

      case 0xD4: goto done;
      case 0xD5: goto done;
      case 0xD6: valid = false; goto done;
      case 0xD7: goto done;

      case 0xD8: // ESC0
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         if (mod != 3)
         {
            switch (opcode)
            {
            case 0x00: Es(); goto done;
            case 0x01: Es(); goto done;
            case 0x02: Es(); goto done;
            case 0x03: Es(); goto done;
            case 0x04: Es(); goto done;
            case 0x05: Es(); goto done;
            case 0x06: Es(); goto done;
            case 0x07: Es(); goto done;
            default: valid = false; goto done;
            }
         }
         else
         {
            // mod == 3
            switch (opcode)
            {
            case 0x00: ST_STi(); goto done;
            case 0x01: ST_STi(); goto done;
            case 0x02: ST_STi(); goto done;
            case 0x03: ST_STi(); goto done;
            case 0x04: ST_STi(); goto done;
            case 0x05: ST_STi(); goto done;
            case 0x06: ST_STi(); goto done;
            case 0x07: ST_STi(); goto done;
            default: valid = false; goto done;
            }
         }

      case 0xD9: // ESC1
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         if (mod != 3)
         {
            switch (opcode)
            {
            case 0x00: Es(); goto done;
            case 0x01: valid = false; goto done;
            case 0x02: Es(); goto done;
            case 0x03: Es(); goto done;
            case 0x04: Ea(); goto done;
            case 0x05: Ew(); goto done;
            case 0x06: Ea(); goto done;
            case 0x07: Ew(); goto done;
            default: valid = false; goto done;
            }
         }
         else
         {
            // mod == 3
            switch (opcode)
            {
            case 0x00:
               STi(); goto done;
            case 0x01:
               STi(); goto done;
            case 0x02:
               if (rm == 0)
                  goto done;
               else
                  valid = false; goto done;
            case 0x03:
               STi(); goto done;
            case 0x04:
               switch (rm)
               {
               case 0x00: goto done;
               case 0x01: goto done;
               case 0x02:
               case 0x03: valid = false; goto done;
               case 0x04: goto done;
               case 0x05: goto done;
               case 0x06:
               case 0x07: valid = false; goto done;
               }
            case 0x05:
               switch (rm)
               {
               case 0x00: goto done;
               case 0x01: goto done;
               case 0x02: goto done;
               case 0x03: goto done;
               case 0x04: goto done;
               case 0x05: goto done;
               case 0x06: goto done;
               case 0x07: valid = false; goto done;
               }
            case 0x06:
               switch (rm)
               {
               case 0x00: goto done;
               case 0x01: goto done;
               case 0x02: goto done;
               case 0x03: goto done;
               case 0x04: goto done;
               case 0x05: goto done;
               case 0x06: goto done;
               case 0x07: goto done;
               }
            case 0x07:
               switch (rm)
               {
               case 0x00: goto done;
               case 0x01: goto done;
               case 0x02: goto done;
               case 0x03: goto done;
               case 0x04: goto done;
               case 0x05: goto done;
               case 0x06: goto done;
               case 0x07: goto done;
               }
            default: valid = false; goto done;
            }
         }

      case 0xDA: // ESC2
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         if (mod != 3)
         {
            switch (opcode)
            {
            case 0x00: Ed(); goto done;
            case 0x01: Ed(); goto done;
            case 0x02: Ed(); goto done;
            case 0x03: Ed(); goto done;
            case 0x04: Ed(); goto done;
            case 0x05: Ed(); goto done;
            case 0x06: Ed(); goto done;
            case 0x07: Ed(); goto done;
            default: valid = false; goto done;
            }
         }
         else
         {
            // mod == 3
            switch (opcode)
            {
            case 0x05:
               if (rm == 1)
                  goto done;
               else
                  valid = false; goto done;
            default: valid = false;goto done;
            }
         }

      case 0xDB: // ESC3
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         if (mod != 3)
         {
            switch (opcode)
            {
            case 0x00: Ed(); goto done;
            case 0x01: valid = false; goto done;
            case 0x02: Ed(); goto done;
            case 0x03: Ed(); goto done;
            case 0x04: valid = false; goto done;
            case 0x05: Et(); goto done;
            case 0x06: valid = false; goto done;
            case 0x07: Et(); goto done;
            default: valid = false; goto done;
            }
         }
         else
         {
            // mod == 3
            switch (opcode)
            {
            case 0x04:
               switch (rm)
               {
               case 0x00: goto done;
               case 0x01: goto done;
               case 0x02: goto done;
               case 0x03: goto done;
               case 0x04: goto done;
               default: valid = false; goto done;
               }
            default: valid = false; goto done;
            }
         }

      case 0xDC: // ESC4
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         if (mod != 3)
         {
            switch (opcode)
            {
            case 0x00: El(); goto done;
            case 0x01: El(); goto done;
            case 0x02: El(); goto done;
            case 0x03: El(); goto done;
            case 0x04: El(); goto done;
            case 0x05: El(); goto done;
            case 0x06: El(); goto done;
            case 0x07: El(); goto done;
            default: valid = false; goto done;
            }
         }
         else
         {
            // mod == 3
            switch (opcode)
            {
            case 0x00: STi_ST(); goto done;
            case 0x01: STi_ST(); goto done;
            case 0x02: STi_ST(); goto done;
            case 0x03: STi_ST(); goto done;
            case 0x04: STi_ST(); goto done;
            case 0x05: STi_ST(); goto done;
            case 0x06: STi_ST(); goto done;
            case 0x07: STi_ST(); goto done;
            default: valid = false; goto done;
            }
         }

      case 0xDD: /* ESC5 */
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         if (mod != 3)
         {
            switch (opcode)
            {
            case 0x00: El(); goto done;
            case 0x01: valid = false; goto done;
            case 0x02: El(); goto done;
            case 0x03: El(); goto done;
            case 0x04: Ea(); goto done;
            case 0x05: valid = false; goto done;
            case 0x06: Ea(); goto done;
            case 0x07: Ew(); goto done;
            default: valid = false; goto done;
            }
         }
         else
         {
            // mod == 3
            switch (opcode)
            {
            case 0x00: STi(); goto done;
            case 0x01: STi(); goto done;
            case 0x02: STi(); goto done;
            case 0x03: STi(); goto done;
            case 0x04: STi_ST(); goto done;
            case 0x05: STi(); goto done;
            case 0x06: valid = false; goto done;
            case 0x07: valid = false; goto done;
            default: valid = false; goto done;
            }
         }

      case 0xDE: // ESC6
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         if(mod != 3)
         {
            switch (opcode)
            {
            case 0x00: Ew(); goto done;
            case 0x01: Ew(); goto done;
            case 0x02: Ew(); goto done;
            case 0x03: Ew(); goto done;
            case 0x04: Ew(); goto done;
            case 0x05: Ew(); goto done;
            case 0x06: Ew(); goto done;
            case 0x07: Ew(); goto done;
            default: valid = false; goto done;
            }
         }
         else
         {
            // mod == 3
            switch (opcode)
            {
            case 0x00: STi_ST(); goto done;
            case 0x01: STi_ST(); goto done;
            case 0x02: STi(); goto done;
            case 0x03:
               switch (rm)
               {
               case 0x01: goto done;
               default: valid = false; goto done;
               }
            case 0x04: STi_ST(); goto done;
            case 0x05: STi_ST(); goto done;
            case 0x06: STi_ST(); goto done;
            case 0x07: STi_ST(); goto done;
            default: valid = false; goto done;
            }
         }

      case 0xDF: // ESC7
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         if (mod != 3)
         {
            switch (opcode)
            {
            case 0x00: Ew(); goto done;
            case 0x01: valid = false; goto done;
            case 0x02: Ew(); goto done;
            case 0x03: Ew(); goto done;
            case 0x04: Eb(); goto done;
            case 0x05: Eq(); goto done;
            case 0x06: Eb(); goto done;
            case 0x07: Eq(); goto done;
            default: valid = false; goto done;
            }
         }
         else
         { // mod == 3
            switch (opcode)
            {
            case 0x00: STi(); goto done;
            case 0x01: STi(); goto done;
            case 0x02: STi(); goto done;
            case 0x03: STi(); goto done;
            case 0x04:
               switch (rm)
               {
               case 0x00: goto done;
               default: valid = false; goto done;
               }
            default: valid = false; goto done;
            }
         }

      case 0xE0: Jb(); goto done;
      case 0xE1: Jb(); goto done;
      case 0xE2: Jb(); goto done;
      case 0xE3: Jb(); goto done;
      case 0xE4: Ib(); goto done;
      case 0xE5: eAX(); Ib(); goto done;
      case 0xE6: Ib(); goto done;
      case 0xE7: Ib(); eAX(); goto done;
      case 0xE8: Av(); goto done;
      case 0xE9: Jv(); goto done;
      case 0xEA: Ap(); goto done;
      case 0xEB: Jb(); goto done;
      case 0xEC: goto done;
      case 0xED: eAX(); goto done;
      case 0xEE: goto done;
      case 0xEF: eAX(); goto done;

      case 0xF0: // LOCK
         break;
      case 0xF1: valid = false; goto done;
      case 0xF2: // REPNE/REPNZ
         db_repne_prefix = 1;
      break;
      case 0xF3: // REP/REPE/REPZ
         db_rep_prefix = 1;
         break;
      case 0xF4: goto done;
      case 0xF5: goto done;
      case 0xF6: // Group 3 Eb
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         switch (opcode)
         {
         case 0x00: EbIb(); goto done;
         case 0x01: valid = false; goto done;
         case 0x02: Eb(); goto done;
         case 0x03: Eb(); goto done;
         case 0x04: Eb(); goto done;
         case 0x05: Eb(); goto done;
         case 0x06: Eb(); goto done;
         case 0x07: Eb(); goto done;
         default: valid = false; goto done;
         }
      case 0xF7: // GROUP3 Ev
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         switch (opcode)
         {
         case 0x00: EvIv(); goto done;
         case 0x01: valid = false; goto done;
         case 0x02: Ev(); goto done;
         case 0x03: Ev(); goto done;
         case 0x04: eAXEv(); goto done;
         case 0x05: eAXEv(); goto done;
         case 0x06: eAXEv(); goto done;
         case 0x07: eAXEv(); goto done;
         default: valid = false; goto done;
         }
      case 0xF8: goto done;
      case 0xF9: goto done;
      case 0xFA: goto done;
      case 0xFB: goto done;
      case 0xFC: goto done;
      case 0xFD: goto done;
      case 0xFE: // GROUP4
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         switch (opcode)
         {
         case 0x00: Eb(); goto done;
         case 0x01: Eb(); goto done;
         default: valid = false; goto done;
         }

      case 0xFF: // GROUP 5
         mod_rm_byte = peek_byte();
         BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
         switch (opcode)
         {
         case 0x00: Ev(); goto done;
         case 0x01: Ev(); goto done;
         case 0x02: Ev(); goto done;
         case 0x03: Ep(); goto done;
         case 0x04: Ev(); goto done;
         case 0x05: Ep(); goto done;
         case 0x06: Ev(); goto done;
         default: valid = false; goto done;
         }

      default: // only invalid instructions left
         valid = false;
         goto done;
      }
   }
   valid = false;

done:
   length = (instruction - instruction_begin);
   return valid;
}


// instruction accessors
//-----------------------------
unsigned char fetch_byte(void)
{
   return(*instruction++);
}

unsigned char peek_byte(void)
{
   return(*instruction);
}

unsigned short fetch_word(void)
{
   unsigned short ret16;
   unsigned char  b1, b0;

   b0 = * (unsigned char *) instruction++;
   b1 = * (unsigned char *) instruction++;
   ret16 = (b1<<8) | b0;
   return(ret16);
}

unsigned int fetch_dword(void)
{
   unsigned int ret32;
   unsigned char  b3, b2, b1, b0;

   b0 = * (unsigned char *) instruction++;
   b1 = * (unsigned char *) instruction++;
   b2 = * (unsigned char *) instruction++;
   b3 = * (unsigned char *) instruction++;
   ret32 = (b3<<24) | (b2<<16) | (b1<<8) | b0;
   return(ret32);
}


// op methods
//-----------------------------

// Debug, Test, and Control Register stuff
void RdCd(void)
{
   unsigned char modrm;
   modrm = fetch_byte();
   if( (modrm >> 6) != 3 )
      valid=false; 
}

void CdRd(void)
{
   unsigned char modrm;
   modrm = fetch_byte();
   if( (modrm >> 6) != 3 )
      valid=false; 
}

void eAXOv(void)
{
   if(db_32bit_addrsize)
      fetch_dword();
   else
      fetch_word();
}

void OveAX(void)
{
   if(db_32bit_addrsize)
      fetch_dword();
   else
      fetch_word();
}

void XvYv(void)
{
}

void GvEb(void)
{
   if(db_32bit_opsize)
      decode_gxex(BX_GENERAL_32BIT_REG, BX_GENERAL_8BIT_REG);
   else
      decode_gxex(BX_GENERAL_16BIT_REG, BX_GENERAL_8BIT_REG);
}

void Av(void)
{
   if(db_32bit_opsize)
      fetch_dword();
   else
      fetch_word();
}

void Eb(void)
{
  decode_exgx(BX_GENERAL_8BIT_REG, BX_NO_REG_TYPE);
}

void Eb1(void)
{
  decode_exgx(BX_GENERAL_8BIT_REG, BX_NO_REG_TYPE);
}

void Ev1(void)
{
   if (db_32bit_opsize)
      decode_exgx(BX_GENERAL_32BIT_REG, BX_NO_REG_TYPE);
   else
      decode_exgx(BX_GENERAL_16BIT_REG, BX_NO_REG_TYPE);
}

void Iw(void)
{
   fetch_word();
}

void EbGb(void)
{
   decode_exgx(BX_GENERAL_8BIT_REG, BX_GENERAL_8BIT_REG);
}

void EvGv(void)
{
   if(db_32bit_opsize)
      decode_exgx(BX_GENERAL_32BIT_REG, BX_GENERAL_32BIT_REG);
   else
      decode_exgx(BX_GENERAL_16BIT_REG, BX_GENERAL_16BIT_REG);
}

void GbEb(void)
{
   decode_gxex(BX_GENERAL_8BIT_REG, BX_GENERAL_8BIT_REG);
}

void GvEv(void)
{
   if(db_32bit_opsize)
      decode_gxex(BX_GENERAL_32BIT_REG, BX_GENERAL_32BIT_REG);
   else
      decode_gxex(BX_GENERAL_16BIT_REG, BX_GENERAL_16BIT_REG);
}

void Ew(void)
{
   decode_exgx(BX_GENERAL_16BIT_REG, BX_NO_REG_TYPE);
}

void GvEw(void)
{
   if(db_32bit_opsize)
      decode_gxex(BX_GENERAL_32BIT_REG, BX_GENERAL_16BIT_REG);
   else
      decode_gxex(BX_GENERAL_16BIT_REG, BX_GENERAL_16BIT_REG);
}

void Jv(void)
{
   if(db_32bit_opsize)
      fetch_dword();
   else
      fetch_word();
}

void EvIb(void)
{
   if (db_32bit_opsize)
   {
      decode_exgx(BX_GENERAL_32BIT_REG, BX_NO_REG_TYPE);
      fetch_byte();
   }
   else
   {
      decode_exgx(BX_GENERAL_16BIT_REG, BX_NO_REG_TYPE);
      fetch_byte();
   }
}

void Iv(void)
{
   if(db_32bit_opsize)
      fetch_dword();
   else
      fetch_word();
}

void Ib(void)
{
   fetch_byte();
}

void Jb(void)
{
   fetch_byte();
}

void EbIb(void)
{
   decode_exgx(BX_GENERAL_8BIT_REG, BX_NO_REG_TYPE);
   fetch_byte();
}

void EvIv(void)
{
   if (db_32bit_opsize)
   {
      decode_exgx(BX_GENERAL_32BIT_REG, BX_NO_REG_TYPE);
      fetch_dword();
   }
   else
   {
      decode_exgx(BX_GENERAL_16BIT_REG, BX_NO_REG_TYPE);
      fetch_word();
   }
}

void EwSw(void)
{
   decode_exgx(BX_GENERAL_16BIT_REG, BX_SEGMENT_REG);
}

void GvM(void)
{
   if(db_32bit_opsize)
      decode_gxex(BX_GENERAL_32BIT_REG, BX_GENERAL_32BIT_REG);
   else
      decode_gxex(BX_GENERAL_16BIT_REG, BX_GENERAL_16BIT_REG);
}

void SwEw(void)
{
   decode_gxex(BX_SEGMENT_REG, BX_GENERAL_16BIT_REG);
}

void Ev(void)
{
   if(db_32bit_opsize)
      decode_exgx(BX_GENERAL_32BIT_REG, BX_NO_REG_TYPE);
   else
      decode_exgx(BX_GENERAL_16BIT_REG, BX_NO_REG_TYPE);
}

void Ap(void)
{
   if(db_32bit_opsize)
   {
      fetch_dword();
      fetch_word();
   }
   else
   {
      fetch_word();
      fetch_word();
   }
}

void XbYb(void)
{
}

void YveAX(void)
{
}

void GvMp(void)
{
   if (db_32bit_opsize)
      decode_gxex(BX_GENERAL_32BIT_REG, BX_GENERAL_32BIT_REG);
   else
      decode_gxex(BX_GENERAL_16BIT_REG, BX_GENERAL_16BIT_REG);
}

void eAXEv(void)
{
   if (db_32bit_opsize)
      decode_gxex(BX_NO_REG_TYPE, BX_GENERAL_32BIT_REG);
   else
      decode_gxex(BX_NO_REG_TYPE, BX_GENERAL_16BIT_REG);
}

void Ep(void)
{
   if (db_32bit_opsize)
      decode_exgx(BX_GENERAL_32BIT_REG, BX_NO_REG_TYPE);
   else
      decode_exgx(BX_GENERAL_16BIT_REG, BX_NO_REG_TYPE);
}

void eAX(void)
{
}

void eCX(void)
{
}

void eDX(void)
{
}

void eBX(void)
{
}

void eSP(void)
{
}

void eBP(void)
{
}

void eSI(void)
{
}

void eDI(void)
{
}

void decode_gxex(unsigned reg_type, unsigned modrm_reg_type)
{
   decode_ex(fetch_byte(), modrm_reg_type);
}

void decode_ex(unsigned char modrm, unsigned modrm_reg_type)
{
   unsigned char   mod, ttt, rm;

   mod = modrm >> 6;
   ttt = (modrm >> 3) & 0x07;
   rm = modrm & 0x07;

   if(db_32bit_addrsize)
   {
      unsigned char    sib, ss, index, base;

      // use 32bit addressing modes.  orthogonal base & index registers,
      // scaling available, etc.
      if(mod != 3)
      {
         // mod != 3
         if (rm != 4)
         {
            // rm != 100b, no s-i-b byte
            // one byte modrm
            switch (mod)
            {
            case 0:
               if (rm == 5)   // no reg, 32-bit displacement
                  fetch_dword();
               break;
            case 1:
               // reg, 8-bit displacement, sign extend
               fetch_byte();
               break;
            case 2:
               // reg, 32-bit displacement
               fetch_dword();
               break;
            }
         }
         else
         {
            // rm == 4, s-i-b byte follows
            sib = fetch_byte();
            ss = sib >> 6;
            index = (sib >> 3) & 0x07;
            base = sib & 0x07;

            switch (mod)
            {
            case 0:
               if(base == 5)
                  fetch_dword();
               break;
            case 1:
               fetch_byte();
               break;
            case 2:
               fetch_dword();
               break;
            }
         } // s-i-b byte follows
      }
   }
   else
   {
      // 16 bit addressing modes.
      switch (mod)
      {
      case 0:
         switch (rm)
         {
            case 6: // DS:d16
            fetch_word();
            break;
         }
         break;
      case 1:
         fetch_byte();
         break;
      case 2:
         fetch_word();
         break;
      }
   }
}

void decode_exgx(unsigned modrm_reg_type, unsigned reg_type)
{
   decode_ex(fetch_byte(), modrm_reg_type);
}


// TODO
//-----------------------------

// Floating point stuff
void ST_STi(void) { valid=false; }
void STi_ST(void) { valid=false; }
void STi(void) { valid=false; }


// Debug, Test, and Control Register stuff
void RdDd(void) { valid=false; }
void DdRd(void) { valid=false; }
void RdTd(void) { valid=false; }
void TdRd(void) { valid=false; }

// Other un-implemented operand signatures
void Ms(void) { valid=false; }
void XBTS(void) { valid=false; }
void IBTS(void) { valid=false; }
void Mp(void) { valid=false; }
void GvMa(void) { valid=false; }
void EwRw(void) { valid=false; }
void YbDX(void) { valid=false; }
void YvDX(void) { valid=false; }
void DXXb(void) { valid=false; }
void DXXv(void) { valid=false; }
void ALOb(void) { valid=false; }

void ObAL(void) { valid=false;}
void YbAL(void) { valid=false;}
void ALXb(void) { valid=false;}
void eAXXv(void) { valid=false;}
void Es(void) { valid=false;}
void Ea(void) { valid=false;}
void Et(void) { valid=false;}
void Ed(void) { valid=false;}
void El(void) { valid=false;}
void Eq(void) { valid=false;}
